from typing import Dict, Union

from fastapi import APIRouter
import pandas as pd
from pydantic import Field


from model_server.base.accessors import GenericImageDataAccessor, PatchStack
from model_server.base.pipelines.params import PipelineParams, PipelineRecord
from model_server.base.pipelines.util import call_pipeline
from model_server.base.roiset import RoiSet, RoiSetMetaParams, RoiSetExportParams
from model_server.base.session import session
from model_server.base.util import PipelineTrace

from model_server.base.models import Model, InstanceSegmentationModel, SemanticSegmentationModel

router = APIRouter(
    prefix='/pipelines',
)

class RoiSetObjectMapParams(PipelineParams):
    accessor_id: str = Field(
        description='ID(s) of previously loaded accessor(s) to use as pipeline input'
    )
    pixel_classifier_segmentation_model_id: str = Field(
        description='Pixel classifier applied to segmentation_channel(s) to segment objects'
    )
    object_classifier_model_id: Union[str, None] = Field(
        None,
        description='Object classifier used to classify segmented objectss'
    )
    pixel_classifier_derived_model_id: Union[str, None] = Field(
        None,
        description='Pixel classifier used to derive channel(s) as additional inputs to object classification'
    )
    segmentation_channel: int = Field(
        description='Channel of input image to use for solving segmentation'
    )
    patches_channel: int = Field(
        description='Channel of input image used in patches sent to object classifier'
    )
    zmask_zindex: Union[int, None] = Field(
        None,
        description='z coordinate to use on input image when solving segmentation; apply MIP if empty',
    )
    roi_params: RoiSetMetaParams = RoiSetMetaParams(**{
        'mask_type': 'boxes',
        'filters': {
            'area': {'min': 1e3, 'max': 1e8}
        },
        'expand_box_by': [128, 2]
    })
    # TODO: maybe don't support all these exports here; instead leverage interm accessors
    export_params: RoiSetExportParams = RoiSetExportParams(**{
        'annotated_patches_2d': {
            'draw_bounding_box': True,
            'pad_to': 256,
        },
        'patches_2d': {
            'draw_bounding_box': True,
            'draw_mask': False,
        },
        'patch_masks': {},
        'object_classes': True,
        'dataframe': True,
    })
    derived_channels_input_channel: Union[int, None] = Field(
        None,
        description='Channel of input image from which to compute derived channels; use all if empty'
    )
    derived_channels_output_channels: Union[int, list] = Field(
        None,
        description='Derived channels to send to object classifier; use all if empty'
    )
    # TODO: move to subclassed pipeline
    # label_params: Union[LabelFromBoundarySegParams, None] = None
    export_label_interm: bool = False


class RoiSetToObjectMapRecord(PipelineRecord):
    pass

@router.put('/roiset_to_obmap/infer')
def roiset_object_map(p: RoiSetObjectMapParams) -> RoiSetToObjectMapRecord:
    """
    Compute a RoiSet from 2d segmentation, apply to z-stack, and optionally apply object classification.
    """


    record, dataframe = call_pipeline(roiset_object_map_pipeline, p)

    # TODO: try labeling this correctly as input accessor_id instead of filename
    session.write_to_table('RoiSet', {'input_filename': p.accessor_id}, dataframe)
    return record


def roiset_object_map_pipeline(
        accessors: Dict[str, GenericImageDataAccessor],
        models: Dict[str, Model],
        **k
) -> (RoiSetToObjectMapRecord, pd.DataFrame):
    if not isinstance(models['pixel_classifier_segmentation_model'], SemanticSegmentationModel):
        raise IncompatibleModelsError('Expecting a pixel classification model')

    d = PipelineTrace()
    d['raw'] = accessors['accessor']

    # MIP if no zmask z-index is given, then classify pixels
    zmi = k['zmask_zindex']
    sch = k['segmentation_channel']

    if isinstance(zmi, int):
        assert 0 < zmi < d.last.nz
        d['mip'] = d.last.get_mono(channel=sch).get_zi(zmi)
    else:
        d['mip'] = d.last.get_mono(channel=sch).apply(lambda x: x.max(axis=-1, keepdims=True))

    d['mip_mask'] = models['pixel_classifier_segmentation_model'].label_pixel_class(d.last)

    # TODO: subclass this boundary pipeline in separate module
    # # optionally label objects using a boundary segmentation pipeline
    # if p.label_params:
    #     la_nda, interm = label_from_boundary_seg(
    #         mip_mask.data[:, :, 0, 0],
    #         p.label_params,
    #         return_interm=True
    #     )
    #     labels = InMemoryDataAccessor(la_nda)
    #     rois = RoiSet(stack, labels, params=p.roi_params)
    # else:
    #     rois = RoiSet.from_segmentation(stack, mip_mask, params=p.roi_params)
    # ti.click('generate_zmasks')

    rois = RoiSet.from_binary_mask(d['raw'], d.last, RoiSetMetaParams(**k['roi_params']))

    # optionally derive additional inputs for object classification
    if dpmod := models.get('pixel_classifier_derived'):
        ic = k['derived_channels_input_channel']
        ocs = k['derived_channels_output_channels']
        assert ic < d['raw'].chroma
        if not isinstance(dpmod, SemanticSegmentationModel):
            raise IncompatibleModelsError('Expecting pixel_classifier_derived to be a pixel classification model')

        def _derive(acc: GenericImageDataAccessor, oc: int):
            acc_mono = acc.get_channels([ic])
            pxmap = dpmod.infer_patch_stack(acc_mono)[0]
            assert oc < pxmap.chroma
            return PatchStack((pxmap.get_channels([oc]).data * 255).astype('uint8'))
        derived_channel_handles = [
            lambda a: _derive(a, oc) for oc in ocs
        ]
    else:
        derived_channel_handles = None

    # optionally classify if an object classifier is passed
    if obmod := models.get('object_classifier_model'):
        obmod_name = k['object_classifier_model_id']
        assert isinstance(obmod, InstanceSegmentationModel)
        rois.classify_by(
            obmod_name,
            [k['patches_channel']],
            obmod,
            derived_channel_functions=derived_channel_handles
        )
        d[obmod_name] = rois.get_object_class_map(obmod_name)

    # TODO: subclass this boundary pipeline in separate module
    # if p.label_params and p.export_label_interm:
    #     fp_la_interm = Path(output_folder_path) / 'mip_masks' / f'interm_{fstem}.png'
    #     plot_image_sequence_with_markers(interm, fp_la_interm, n_rows=3)
    #     record['label_interm'] = fp_la_interm

    return d

class Error(Exception):
    pass

class IncompatibleModelsError(Error):
    pass