import shutil
from pathlib import Path
from shutil import copyfile
import unittest

import numpy as np

import model_server.conf.testing as conf

from model_server.base.accessors import generate_file_accessor
from tests.base.test_model import DummyInstanceSegmentationModel
from model_server.base.roiset import RoiSetMetaParams, RoiSetExportParams

import model_server.conf.testing as conf
# from extensions.chaeo.workflows import ClassifyZStackApiParams, export_zstack_roiset
from model_server.base.pipelines.roiset_obmap import RoiSetObjectMapParams, roiset_object_map_pipeline
from model_server.extensions.ilastik.models import IlastikPixelClassifierModel, IlastikPixelClassifierParams

data = conf.meta['image_files']
output_path = conf.meta['output_path']
test_params = conf.meta['roiset']
classifiers = conf.meta['ilastik_classifiers']

class BaseTestRoiSetMonoProducts(object):

    def setUp(self) -> None:
        # set up test raw data and segmentation from file
        self.stack = generate_file_accessor(data['multichannel_zstack_raw']['path'])
        self.stack_ch_pa = self.stack.get_mono(test_params['patches_channel'])
        self.seg_mask = generate_file_accessor(data['multichannel_zstack_mask2d']['path'])


def _get_export_params():
    return RoiSetExportParams(**{
        'pixel_probabilities': True,
        'patches_3d': {},
        'annotated_patches_2d': {
            'draw_bounding_box': True,
            'rgb_overlay_channels': [3, None, None],
            'rgb_overlay_weights': [0.2, 1.0, 1.0],
            'pad_to': 512,
        },
        'patches_2d': {
            'draw_bounding_box': False,
            'draw_mask': False,
        },
        'patch_masks': {
            'pad_to': 256,
        },
        'annotated_zstacks': {},
        'object_classes': True,
        'dataframe': True,
    })

def _get_roi_params():
    return RoiSetMetaParams(**{
        'mask_type': 'boxes',
        'filters': {
            'area': {'min': 1e3, 'max': 1e8}
        },
        'expand_box_by': [128, 2]
    })

# def _get_model_params():
#     return {
#         'pixel_classifier_segmentation': {
#             # 'project_file': params['pixel_classifier'],
#             'project_file': classifiers['px']['path'],
#         },
#         'object_classifier': {
#             'name': 'dummy',
#         }
#     }

def _get_models():

    return {
        'pixel_classifier_segmentation': {
            'name': 'ilastik_px_mod',
            'project_file': classifiers['px']['path'],
            'model': IlastikPixelClassifierModel(
                IlastikPixelClassifierParams(
                    project_file=classifiers['px']['path'].__str__()
                )
            )
        },
        'object_classifier': {
            'name': 'dummy_ob_mod',
            'model': DummyInstanceSegmentationModel()
        },

    }


class TestRoiSetWorkflow(BaseTestRoiSetMonoProducts, unittest.TestCase):

    def test_object_map_workflow(self):
        acc_in = generate_file_accessor(data['multichannel_zstack_raw']['path'])
        params = RoiSetObjectMapParams(
            api=False,
            accessor_id='acc_id',
            pixel_classifier_segmentation_model_id='px_id',
            object_classifier_model_id='ob_id',
            segmentation_channel=test_params['segmentation_channel'],
            patches_channel=test_params['patches_channel'],
            roi_params=_get_roi_params(),
            export_params=_get_export_params(),
        )
        trace = roiset_object_map_pipeline(
            {'accessor': acc_in},
            {f'{k}_model': v['model'] for k, v in _get_models().items()},
            **params.dict()
        )

        self.assertTrue('ob_id' in trace.keys())
        self.assertEqual(len(trace['ob_id']._unique()[0]), 2)


class TestRoiSetWorkflowOverApi(conf.TestServerBaseClass):

    app_name = 'api:app'

    def _copy_input_file_to_server(self):
        resp = self._get('paths')
        pa = resp.json()['inbound_images']
        copyfile(
            data['multichannel_zstack_raw']['path'],
            Path(pa) / data['multichannel_zstack_raw']['name']
        )

    def setUp(self) -> None:
        self.where_out = output_path / 'trec-adaptive-feedback' / 'roiset'
        self.where_out.mkdir(parents=True, exist_ok=True)
        return super().setUp()

    def test_bounceback(self):
        resp = self._put('chaeo/bounce_back', query={'par1': 'hello'})
        self.assertEqual(resp.status_code, 200, resp.json())
        self.assertEqual(resp.json()['params']['par1'], 'hello', resp.json())

    def test_load_pixel_classifier(self):
        resp = self._put(
            'ilastik/seg/load/',
            body={'project_file': _get_model_params()['pixel_classifier_segmentation']['project_file'].__str__()},
        )
        model_id = resp.json()['model_id']
        self.assertTrue(model_id.startswith('IlastikPixelClassifierModel'))
        return model_id

    def test_load_object_classifier(self):
        resp = self._put(f'models/dummy_instance/load')
        model_id = resp.json()['model_id']
        self.assertTrue(model_id.startswith('DummyInstanceSegmentationModel'))
        return model_id

    def test_object_map_workflow(self):
        mid_px = self.test_load_pixel_classifier()
        mid_ob = self.test_load_object_classifier()
        resp = self._put(
            'chaeo/classify_zstack/infer',
            body=ClassifyZStackApiParams(
                model_ids={
                    'pixel_classifier_segmentation': mid_px,
                    'object_classifier': mid_ob,
                },
                **{
                    'input_filename': data['multichannel_zstack_raw']['path'].__str__(),
                    'segmentation_channel': 0,
                    'patches_channel': 1,
                    'roi_params': _get_roi_params(),
                    'export_params': {'object_classes': True},
                },
            ).dict()
        )
        self.assertEqual(resp.status_code, 200, resp.json())
        omfp = Path(resp.json()['object_map_filepath'])
        self.assertTrue(omfp.exists())
        acc_obmap = generate_file_accessor(omfp)
        self.assertTrue(all(np.unique(acc_obmap.data) == [0, 1]))
        shutil.copy(omfp, self.where_out / f'normal_{omfp.name}')

    def test_workflow_without_object_classifier(self):
        mid_px = self.test_load_pixel_classifier()
        resp = self._put(
            'chaeo/classify_zstack/infer',
            body=ClassifyZStackApiParams(
                model_ids={
                    'pixel_classifier_segmentation': mid_px,
                },
                **{
                    'input_filename': data['multichannel_zstack_raw']['path'].__str__(),
                    'segmentation_channel': 0,
                    'patches_channel': 1,
                    'roi_params': _get_roi_params(),
                    'export_params': {'object_classes': True},
                },
            ).dict()
        )
        self.assertEqual(resp.status_code, 200, resp.json())
        omfp = Path(resp.json()['object_map_filepath'])
        self.assertTrue(omfp.exists())
        acc_obmap = generate_file_accessor(omfp)
        self.assertTrue(all(np.unique(acc_obmap.data) == [0, 1]))
        shutil.copy(omfp, self.where_out / f'normal_{omfp.name}')

    def test_object_map_workflow_boundary_channel(self):
        fp_in = data['multichannel_zstack_raw']['path']
        resp = self._put(
            'ilastik/seg/load/',
            body={
                'project_file': classifiers['px']['path'].__str__(),
                'px_class': 1,
                'px_prob_threshold': 0.5
            },
        )
        mid_px = resp.json()['model_id']
        mid_ob = self.test_load_object_classifier()
        resp = self._put(
            'chaeo/classify_zstack/infer',
            body=ClassifyZStackApiParams(
                model_ids={
                    'pixel_classifier_segmentation': mid_px,
                    'object_classifier': mid_ob,
                },
                **{
                    'input_filename': fp_in.__str__(),
                    'segmentation_channel': 0,
                    'patches_channel': 1,
                    'roi_params': _get_roi_params(),
                    'export_params': {'object_classes': True},
                    'label_params': {
                        'bbox_filter': {'area': {'min': 1e3, 'max': 1e8}},
                        'px_expand': 0,
                        'n_dilate': 2,
                        'marker_source': 'distmax',
                        'min_marker_dist': 10,
                        'background_method': 'threshold',
                        'coarse_sig': 15,
                        'background_tr': 0.2,
                    },
                    'export_label_interm': False,
                },
            ).dict()
        )
        self.assertEqual(resp.status_code, 200, resp.json())
        fp_obmap = Path(resp.json()['object_map_filepath'])
        print(fp_obmap)
        self.assertTrue(fp_obmap.exists())
        acc_obmap = generate_file_accessor(fp_obmap)
        self.assertTrue(all(np.unique(acc_obmap.data) == [0, 1]))



class TestRoiSetWithDerivedChannels(conf.TestServerBaseClass):

    app_name = 'api:app'

    def test_object_map_workflow_with_derived_channels(self):
        models = _get_models()
        models['pixel_classifier_derived'] = models['pixel_classifier_segmentation'] # re-use same classifier
        where_out = output_path / 'roiset' / 'workflow'

        p = ClassifyZStackApiParams(
            input_filename=str(data['multichannel_zstack_raw']['path']),
            model_ids={
                'pixel_classifier_segmentation': 'id_px',
                'object_classifier': 'id_ob',
                'pixel_classifier_derived': 'id_px',
            },
            derived_channels_input_channel=0,
            derived_channels_output_channels=[1],
            segmentation_channel=test_params['segmentation_channel'],
            patches_channel=test_params['patches_channel'],
            roi_params=_get_roi_params(),
            export_params=_get_export_params(),
        )

        record = export_zstack_roiset(
            data['multichannel_zstack_raw']['path'],
            where_out,
            models,
            p
        )
        self.assertTrue(all([Path(pa).exists() for pa in record['interm']['derived_channels']]))

    def test_derived_channels_api(self):
        resp = self._put(
            'ilastik/seg/load/',
            body={'project_file': _get_model_params()['pixel_classifier_segmentation']['project_file'].__str__()},
        )
        self.assertEqual(resp.status_code, 200)
        mid_px = resp.json()['model_id']
        resp = self._put(f'models/dummy_instance/load')
        mid_ob = resp.json()['model_id']


        p = ClassifyZStackApiParams(
            input_filename=data['multichannel_zstack_raw']['path'].__str__(),
            model_ids={
                'pixel_classifier_segmentation': mid_px,
                'object_classifier': mid_ob,
                'pixel_classifier_derived': mid_px,
            },
            derived_channels_input_channel=0,
            derived_channels_output_channels=[1],
            segmentation_channel=test_params['segmentation_channel'],
            patches_channel=test_params['patches_channel'],
            roi_params=_get_roi_params(),
            export_params=_get_export_params(),
        )
        resp = self._put('chaeo/classify_zstack/infer', body=p.dict())
        self.assertEqual(resp.status_code, 200, resp.json())
        omfp = resp.json()['object_map_filepath']
        self.assertTrue(Path(omfp).exists())
        acc_obmap = generate_file_accessor(omfp)
        self.assertTrue(all(np.unique(acc_obmap.data) == [0, 1]))

